"""
:author:

Amine SEHILI <amine.sehili@gmail.com>
2015-2026

:License:

This package is published under the MIT license.
"""

from .audio import *
from .core import *
from .exceptions import *
from .io import *
from .util import *

__version__ = "0.4.0"
